/*
 * Copyright (c) 2001-2002 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 * 
 *	Ohio Trollius
 *	Copyright 1997 The Ohio State University
 *	RBD/GDB
 *
 *	$Id: lamboot.c,v 6.27.2.4 2002/10/09 19:49:58 brbarret Exp $
 *
 *	Function:	- fully-connected LAM booting tool
 *			- boots Trollius on a network of UNIX hosts
 *			- hosts specified using host file syntax
 *			- uses hboot
 */

#include <lam_config.h>

#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <args.h>
#include <debug.h>
#include <kio.h>
#include <lamnet.h>
#include <net.h>
#include <sfh.h>
#include <terror.h>
#include <etc_misc.h>
#include <mpisys.h>
#include <boot.h>

/*
 * static variables
 */
static char *bhost;	/* boot schema file */
static int fl_debug;	/* debug mode */
static int fl_verbose;	/* verbose mode */
static int nboot;	/* # nodes booted */
static int nrun;	/* # nodes running */


/* 
 * private functions
 */
static char *findbhost(int cmdc, char **cmdv, char **env);
static void bail(int);
static void wipe(void);


int
main(int argc, char *argv[], char *env[])
{
	FILE		*fp;		/* boot schema file pointer */
	int		iorigin;	/* index of origin node */
	int		badhost;	/* bad host index */
	int		nlamnet;	/* lamnet size */
	struct lamnode	*lamnet;	/* network description array */
	int             j;

/*
 * Parse the command line.
 */
	validopts("bdhsvxHV");

	if (do_args(&argc, argv)) {
	  show_help("lamboot", "usage", NULL);
	  exit(EUSAGE);
	}

	if (opt_taken('V')) {
	  lam_show_version(1);
	  exit(0);
	}

	/* Ensure that we are not root */

	if (getuid() == 0 || geteuid() == 0) {
	  show_help(NULL, "deny-root", NULL);
	  exit(EACCES);
	}

	if ((errno = (argc <= 2) ? 0 : EUSAGE)) {
	  show_help("lamboot", "usage", NULL);
	  exit(errno);
	}
 
	if (opt_taken('h')) {
	  show_help("lamboot", "usage", NULL);
	  exit(0);
	}

	fl_debug = opt_taken('d');
	fl_verbose = opt_taken('v') || fl_debug;

	if (! opt_taken('H'))
	  lam_show_version(0);
/*
 * Locate the system file.
 */
	bhost = findbhost(argc, argv, env);
/*
 * Open the system file.
 */
	DBUG("lamboot: opening hostfile %s\n", bhost);
	fp = fopen(bhost, "r");
	if (fp == 0) {
	  show_help("boot", "open-hostfile", "lamboot", bhost, NULL);
	  exit(errno);
	}
/*
 * Parse the system file.
 */
	if (bhostparse(fp, &lamnet, &nlamnet)) {
	  show_help("boot", "bhostparse", "lamboot", bhost, NULL);
	  exit(errno);
	}
	if (fl_debug) {
	  printf("lamboot: found the following hosts:\n");
	  for (j = 0; j < nlamnet; j++)
	    printf("lamboot:   n%d %s\n", j, lamnet[j].lnd_hname);
	}
/*
 * Close the system file.
 */
	if (fclose(fp)) {
	  show_help("boot", "close-hostfile", "lamboot", bhost, NULL);
	  exit(errno);
	}
/*
 * Locate the host nodes.
 */
	if (lamnet_findhosts(lamnet, nlamnet, &badhost)) {

		if (errno == EBADHOST) {
		  show_help("boot", "resolv-hostname", "lamboot",
			    lamnet[badhost].lnd_hname, bhost, NULL);
		  exit(LAM_EEXIT);
		} else {
		  char buffer[16];
		  sprintf(buffer, "%d", errno);
		  show_help("boot", "resolv-unknown", "lamboot", bhost,
			    buffer, NULL);
		}
		exit(errno);
	}
	if (fl_debug) {
	  printf("lamboot: resolved hosts:\n");
	  for (j = 0; j < nlamnet; j++)
	    printf("lamboot:   n%d %s --> %s\n", j, lamnet[j].lnd_hname,
		   inet_ntoa(lamnet[j].lnd_addr.sin_addr));
	}
/*
 * Find the origin node.
 */
	iorigin = lamnet_findorig(lamnet, nlamnet);

	if (iorigin < 0) {
	  show_help("boot", "no-localhost", "lamboot", bhost, NULL);
	  exit(EINVAL);
	}

	_kio.ki_origin = iorigin;
	lamnet[iorigin].lnd_type |= NT_ORIGIN | NT_ME;

	DBUG("lamboot: found %d host node(s)\n", nlamnet);
	DBUG("lamboot: origin node is %d (%s)\n", iorigin,
	     lamnet[iorigin].lnd_hname);

	/* If there is more than one node in the universe, ensure that
	 * the origin node does not resolve to 127.0.0.1, or the
	 * remote nodes will NOT be send info back to lamboot 
	 */
	if (nlamnet > 1) {
	  if (strcmp("127.0.0.1", 
		     inet_ntoa(lamnet[iorigin].lnd_addr.sin_addr)) == 0) {
	    show_help("boot", "resolve-localhost", 
		      "lamboot", lamnet[iorigin].lnd_hname);
	    exit(EINVAL);
	  }
	}

/*
 * Clean up on interrupt.
 */
	if (signal(SIGINT, bail) == SIG_ERR) 
	  lampanic("lamboot (signal)");
/*
 * Boot the system.
 */
	if (lambootagent(lamnet, nlamnet, &nboot, &nrun)) {

	  /* If the localhost was not in the boot schema, we must
             print the error message here, because lambootagent() does
             not know the name of the file (which is printed in the
             help message) */
	  /* This shouldn't happen, though, since we check for the
             local host above.  But better safe than sorry... */

	  if (errno == EINVAL)
	    show_help("boot", "no-localhost", "lamboot", bhost, NULL);

	  /* lambootagent should print all other relevant error messages */
	  
	  bail(1);
	}

	DBUG("lamboot completed successfully\n");
	return(0);
}

/*
 *	findbhost
 *
 *	Function:	- locates boot schema file
 *	Accepts:	- argc of remaining command line
 *			- argv of remaining command line
 *	Returns:	- full pathname of boot schema file
 */
static char *
findbhost(int cmdc, char **cmdv, char **env)
{
	char		*fname;
	char		*full;		/* temporary for full pathname */
	char		**pathv;
	int		pathc;
/*
 * Set the directories for the boot schema file search.
 */
	pathc = 0;
	pathv = 0;
	argvadd(&pathc, &pathv, "");
	argvadd(&pathc, &pathv, "$TROLLIUSHOME/etc");
	argvadd(&pathc, &pathv, "$LAMHOME/etc");
	argvadd(&pathc, &pathv, LAM_SYSCONFDIR);
/*
 * Set the boot schema file names.
 */
	if (cmdc == 2) {
		fname = cmdv[1];
	} else if ((fname = getenv("LAMBHOST"))) {
	} else if ((fname = getenv("TROLLIUSBHOST"))) {
	} else {
		fname = DEFFBHOST;
	}
/*
 * Search the directories.
 */
	full = sfh_path_findv(fname, pathv, R_OK, env);

	if (full == 0) {
	  show_help("boot", "open-hostfile", "lamboot", fname, NULL);
	  exit(errno);
	}

	DBUG("lamboot: boot schema file: %s\n", full);
	return(full);
}

/*
 *	bail
 *
 *	Function:	- cleans up and bails out
 *	Returns:	- does not return, exits with error code
 */
static void
bail(int sig)
{
	int		err_save;	/* saved error code */

	err_save = errno;
	show_help("lamboot", "about-to-wipe", NULL);
	wipe();
	DBUG("lamboot did NOT complete successfully\n");
	exit(err_save);
}

/*
 *	wipe
 *
 *	Function:	- executes the wipe tool
 *			- kills all host Trollius sessions
 */
static void
wipe(void)
{
	int		cmdn;
	int		r;
	char		**cmdv;
	char		buf[16];

	if (nboot <= 0) {
		return;
	}

	cmdn = 0;
	cmdv = 0;
	argvadd(&cmdn, &cmdv, DEFTWIPE);

	if (fl_verbose) {
		argvadd(&cmdn, &cmdv, "-v");
	}

	argvadd(&cmdn, &cmdv, "-n");
	sprintf(buf, "%d", nboot);
	argvadd(&cmdn, &cmdv, buf);

	argvadd(&cmdn, &cmdv, bhost);

	VERBOSE("%s ...\n", DEFTWIPE);

	r = _lam_few(cmdv);

	if (r) {
		errno = r;
		show_help("boot", "wipe-fail", NULL);
		exit(errno);
	}
}
